from holidata.holiday import Region
from holidata.utils import day


class MD(Region):
    def __init__(self, country):
        super().__init__("MD", country)

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on(month=1, day=7) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Traslado de San José") \
            .in_years([2013]) \
            .on(month=3, day=18) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a San José") \
            .in_years([2017, 2023]) \
            .on(month=3, day=20) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Fiesta de la Comunidad de Madrid") \
            .except_for([2016, 2021]) \
            .on(month=5, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Comunidad de Madrid") \
            .in_years([2021]) \
            .on(month=5, day=3) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Corpus Christi") \
            .in_years([2011, 2014, 2015]) \
            .on(day(60).after(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("San José") \
            .in_years([2012, 2015, 2021, 2023]) \
            .on(month=3, day=19) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Trabajo") \
            .in_years([2016]) \
            .on(month=5, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Santiago Apóstol") \
            .in_years([2011, 2016, 2022]) \
            .on(month=7, day=25) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Santiago Apóstol / Día Nacional de Galicia") \
            .in_years([2024]) \
            .on(month=7, day=25) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a Todos los Santos") \
            .in_years([2020]) \
            .on(month=11, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2020]) \
            .on(month=12, day=7) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a La Inmaculada Concepción") \
            .in_years([2019]) \
            .on(month=12, day=9) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on(month=12, day=26) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2016]) \
            .on(month=12, day=26) \
            .with_flags("RF")



from holidata.holiday import Region
from holidata.utils import day


class GA(Region):
    def __init__(self, country):
        super().__init__("GA", country)

        self.define_holiday() \
            .with_name("San Juan") \
            .in_years([2013, 2016, 2020, 2022]) \
            .on(month=6, day=24) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San José") \
            .in_years([2011, 2019, 2020, 2021]) \
            .on(month=3, day=19) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Día siguiente a San José") \
            .in_years([2015]) \
            .on(month=3, day=20) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Día de las Letras Gallegas") \
            .except_for([2015, 2020]) \
            .on(month=5, day=17) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Santiago Apóstol / Día Nacional de Galicia") \
            .except_for([2021]) \
            .on(month=7, day=25) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a Todos los Santos") \
            .in_years([2015]) \
            .on(month=11, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")



from holidata.holiday import Region


class MV(Region):
    def __init__(self, country):
        super().__init__("MV", country)

        self.define_holiday() \
            .with_name("Reformationstag") \
            .until(2016) \
            .on(month=10, day=31) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Reformationstag") \
            .since(2018) \
            .on(month=10, day=31) \
            .with_flags("RF")



from holidata.holiday import Region
from holidata.utils import day


class IB(Region):
    def __init__(self, country):
        super().__init__("IB", country)

        self.define_holiday() \
            .with_name("Día de las Illes Balears") \
            .in_years([2011, 2012, 2013, 2014, 2016, 2017, 2018, 2019, 2021, 2022, 2023, 2024]) \
            .on(month=3, day=1) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes de Pascua") \
            .except_for([2014]) \
            .on(day(1).after(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a Todos los Santos") \
            .in_years([2015]) \
            .on(month=11, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2015, 2020]) \
            .on(month=12, day=7) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on(month=12, day=26) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2011, 2013, 2014, 2016, 2019, 2020]) \
            .on(month=12, day=26) \
            .with_flags("RF")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class BE(Country):
    id = "BE"
    languages = ["de", "fr", "nl"]
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_names({
                "de": "Neujahr",
                "fr": "Nouvel An",
                "nl": "Nieuwjaar",
            }) \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "de": "Tag der Arbeit",
                "fr": "Fête du Travail",
                "nl": "Dag van de arbeid",
            }) \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "de": "Nationalfeiertag",
                "fr": "Fête nationale",
                "nl": "Nationale feestdag",
            }) \
            .on(month=7, day=21) \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "de": "Mariä Himmelfahrt",
                "fr": "Assomption",
                "nl": "Onze Lieve Vrouw hemelvaart",
            }) \
            .on(month=8, day=15) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "de": "Allerheiligen",
                "fr": "Toussaint",
                "nl": "Allerheiligen",
            }) \
            .on(month=11, day=1) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "de": "Waffenstillstand",
                "fr": "Jour de l'armistice",
                "nl": "Wapenstilstand",
            }) \
            .on(month=11, day=11) \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "de": "Weihnacht",
                "fr": "Noël",
                "nl": "Kerstmis",
            }) \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "de": "Ostern",
                "fr": "Pâques",
                "nl": "Pasen",
            }) \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "de": "Ostermontag",
                "fr": "Lundi de Pâques",
                "nl": "Paasmaandag",
            }) \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "de": "Christi Himmelfahrt",
                "fr": "Ascension",
                "nl": "Onze Lieve Heer hemelvaart",
            }) \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "de": "Pfingsten",
                "fr": "Pentecôte",
                "nl": "Pinksteren",
            }) \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "de": "Pfingstmontag",
                "fr": "Lundi de Pentecôte",
                "nl": "Pinkstermaandag",
            }) \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day

"""
sources:
- Hungarian Constitution - Article J (1)
  https://www.keh.hu/magyarorszag_alaptorvenye/1515-Magyarorszag_Alaptorvenye
- 2017. évi XIII. törvény egyes törvényeknek a nagypéntek munkaszüneti nappá történő nyilvánításával összefüggő módosításáról
  https://mkogy.jogtar.hu/jogszabaly?docid=A1700013.TV
"""


class HU(Country):
    id = "HU"
    languages = ["hu"]
    default_lang = "hu"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Újév") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Az 1848-as forradalom ünnepe") \
            .on(month=3, day=15) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("A munka ünnepe") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Az államalapítás ünnepe") \
            .on(month=8, day=20) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Az 1956-os forradalom ünnepe") \
            .on(month=10, day=23) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Mindenszentek") \
            .on(month=11, day=1) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Karácsony") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Karácsony") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Húsvét") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Húsvéthétfő") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pünkösd") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pünkösdhétfő") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Nagypéntek") \
            .since(2017) \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        for year, listing in HU.holiday_munkaszuneti_nap().items():
            for date, note in listing:
                self.define_holiday() \
                    .with_name("Munkaszüneti nap") \
                    .in_years([year]) \
                    .on(**date) \
                    .with_flags("NF") \
                    .annotated_with(note)

    @staticmethod
    def holiday_munkaszuneti_nap():
        """
        Non-Working days (Munkaszüneti nap)
        When a public holiday falls on a Tuesday or a Thursday, a special decree swaps the preceding Monday or the
        following Friday (respectively) with a not too distant Saturday
        2015: 28/2014. (IX.   24.) NGM rendelet a 2015. évi munkaszüneti napok körüli munkarendről
        2016:  8/2015. (VI.   29.) NGM rendelet a 2016. évi munkaszüneti napok körüli munkarendről
        2018:  9/2017. (V.    19.) NGM rendelet a 2018. évi munkaszüneti napok körüli munkarendről
        2019:  6/2018. (VIII. 23.) PM  rendelet a 2019. évi munkaszüneti napok körüli munkarendről
        2020:  7/2019. (VI.   25.) PM  rendelet a 2020. évi munkaszüneti napok körüli munkarendről
        2021: 14/2020. (V.    13.) ITM rendelet a 2021. évi munkaszüneti napok körüli munkarendről
        2022: 23/2021. (VI.    1.) ITM rendelet a 2022. évi munkaszüneti napok körüli munkarendről
        2024: 15/2023. (VII.  13.) GFM rendelet a 2024. évi munkaszüneti napok körüli munkarendről
        """
        return {
            2015: [
                ({"month": 1, "day": 2}, "2015-01-10 munkanap"),
                ({"month": 8, "day": 21}, "2015-08-08 munkanap"),
                ({"month": 12, "day": 24}, "2015-12-12 munkanap"),
            ],
            2016: [
                ({"month": 3, "day": 14}, "2016-03-05 munkanap"),
                ({"month": 10, "day": 31}, "2016-10-15 munkanap"),
            ],
            2018: [
                ({"month": 3, "day": 16}, "2018-03-10 munkanap"),
                ({"month": 4, "day": 30}, "2018-04-21 munkanap"),
                ({"month": 10, "day": 22}, "2018-10-13 munkanap"),
                ({"month": 11, "day": 2}, "2018-11-10 munkanap"),
                ({"month": 12, "day": 24}, "2018-12-01 munkanap"),
                ({"month": 12, "day": 31}, "2018-12-15 munkanap"),
            ],
            2019: [
                ({"month": 8, "day": 19}, "2019-08-10 munkanap"),
                ({"month": 12, "day": 24}, "2019-12-07 munkanap"),
                ({"month": 12, "day": 27}, "2019-12-14 munkanap"),
            ],
            2020: [
                ({"month": 8, "day": 21}, "2020-08-29 munkanap"),
                ({"month": 12, "day": 24}, "2020-12-12 munkanap"),
            ],
            2021: [
                ({"month": 12, "day": 24}, "2021-12-11 munkanap")
            ],
            2022: [
                ({"month": 3, "day": 14}, "2022-03-26 munkanap"),
                ({"month": 10, "day": 31}, "2022-10-15 munkanap"),
            ],
            2024: [
                ({"month": 8, "day": 19}, "2024-08-03 munkanap"),
                ({"month": 12, "day": 24}, "2024-12-07 munkanap"),
                ({"month": 12, "day": 27}, "2024-12-14 munkanap"),
            ],
        }



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class CH(Country):
    id = "CH"
    languages = ["de"]
    default_lang = "de"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Neujahrstag") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Berchtoldstag") \
            .in_regions(["BE", "JU", "TG", "VD"]) \
            .on(month=1, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Heilige Drei Könige") \
            .in_regions(["SZ", "TI", "UR"]) \
            .on(month=1, day=6) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Josefstag") \
            .in_regions(["NW", "SZ", "TI", "UR", "VS"]) \
            .on(month=3, day=19) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Tag der Arbeit") \
            .in_regions(["BL", "BS", "GR", "NE", "SH", "TG", "TI", "ZH"]) \
            .on(month=5, day=1) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Bundesfeier") \
            .on(month=8, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Mariä Himmelfahrt") \
            .in_regions(["AI"]) \
            .on(month=8, day=15) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Mariä Himmelfahrt") \
            .in_regions(["JU", "LU", "NW", "OW", "SZ", "TI", "UR", "VS", "ZG"]) \
            .on(month=8, day=15) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Allerheiligen") \
            .in_regions(["AI", "GL", "JU", "LU", "NW", "OW", "SG", "SZ", "TI", "UR", "VS", "ZG"]) \
            .on(month=11, day=1) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Mariä Empfängnis") \
            .in_regions(["AI"]) \
            .on(month=12, day=8) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Mariä Empfängnis") \
            .in_regions(["LU", "NW", "OW", "SZ", "TI", "UR", "VS", "ZG"]) \
            .on(month=12, day=8) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Weihnachtstag") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Stephanstag") \
            .in_regions(["AI", "AR", "BE", "BL", "BS", "GL", "GR", "LU", "SG", "SH", "SZ", "TG", "TI", "UR", "ZH"]) \
            .on(month=12, day=26) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Karfreitag") \
            .in_regions(["AG", "AI", "AR", "BE", "BL", "BS", "FR", "GE", "GL", "GR", "JU", "LU", "NE", "NW", "OW", "SG", "SH", "SO", "SZ", "TG", "UR", "VD", "ZG", "ZH"]) \
            .on(day(2).before(self.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Ostersonntag") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ostermontag") \
            .in_regions(["AI", "AR", "BE", "BL", "BS", "GE", "GL", "GR", "JU", "SG", "SH", "SZ", "TG", "TI", "UR", "VD", "ZH"]) \
            .on(day(1).after(self.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Auffahrt") \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstsonntag") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstmontag") \
            .in_regions(["AI", "AR", "BE", "BL", "BS", "GE", "GL", "GR", "JU", "SG", "SH", "SZ", "TG", "TI", "UR", "VD", "ZH"]) \
            .on(day(50).after(self.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Fronleichnam") \
            .in_regions(["AI", "JU", "LU", "NW", "OW", "SZ", "TI", "UR", "VS", "ZG"]) \
            .on(day(60).after(self.easter())) \
            .with_flags("RV")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import first, third, last, second, fourth, day


class US(Country):
    id = "US"
    languages = ["en", "es"]
    default_lang = "en"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_names({
                "en": "New Year's Day",
                "es": "Año Neuvo",
            }) \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "en": "Independence Day",
                "es": "Día de la Independiencia",
            }) \
            .on(month=7, day=4) \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "en": "Veterans Day",
                "es": "Día de los Veteranos",
            }) \
            .on(month=11, day=11) \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "en": "Christmas Eve",
                "es": "Nochebuena",
            }) \
            .on(month=12, day=24) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "en": "Christmas Day",
                "es": "Navidad",
            }) \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "en": "Birthday of Martin Luther King, Jr.",
                "es": "Cumpleaños de Martin Luther King, Jr.",
            }) \
            .on(third("monday").of("January")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Washington's Birthday",
                "es": "Día del Presidente",
            }) \
            .on(third("monday").of("february")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Patriots' Day",
                "es": "Día del Patriota",
                }) \
            .in_regions(["MA", "ME"]) \
            .on(third("monday").of("april")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "Memorial Day",
                "es": "Día de los Caídos",
            }) \
            .on(last("monday").of("may")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Labor Day",
                "es": "Día del Trabajo",
            }) \
            .on(first("monday").of("september")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Columbus Day",
                "es": "Día de Columbus",
            }) \
            .on(second("monday").of("october")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Thanksgiving Day",
                "es": "Día de Acción de Gracias",
            }) \
            .on(fourth("thursday").of("november")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Day after Thanksgiving",
                "es": "Día después de Acción de Gracias",
            }) \
            .on(day(1).after(fourth("thursday").of("november"))) \
            .with_flags("NV")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day

"""
source: https://www.zakonyprolidi.cz/cs/2000-245, §1 and §2.
        https://www.zakonyprolidi.cz/cs/2000-245/zneni-20200201 (>2020-02-01)
        https://www.zakonyprolidi.cz/cs/2000-245/zneni-20190401 (>2019-04-01)
        https://www.zakonyprolidi.cz/cs/2000-245/zneni-0 (>2000-08-09)
"""


class CZ(Country):
    id = "CZ"
    languages = ["cs"]
    default_lang = "cs"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nový rok") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Den obnovy samostatného českého státu") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Svátek práce") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Den vítězství") \
            .on(month=5, day=8) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Den slovanských věrozvěstů Cyrila a Metoděje") \
            .on(month=7, day=5) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Den upálení mistra Jana Husa") \
            .on(month=7, day=6) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Den české státnosti") \
            .on(month=9, day=28) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Den vzniku samostatného československého státu") \
            .on(month=10, day=28) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Štědrý den") \
            .on(month=12, day=24) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("1. svátek vánoční") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("2. svátek vánoční") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Velikonoční pondělí") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Velký pátek") \
            .on(day(2).before(self.easter())) \
            .since(2016) \
            .with_flags("NRV")

        """
        Until 2018, replaced by "Den boje za svobodu a demokracii a Mezinárodní den studentstva" on 2019-04-01"
        """
        self.define_holiday() \
            .with_name("Den boje za svobodu a demokracii") \
            .on(month=11, day=17) \
            .until(2018) \
            .with_flags("NF")

        """
        Since 2019, replaces "Den boje za svobodu a demokracii"
        """
        self.define_holiday() \
            .with_name("Den boje za svobodu a demokracii a Mezinárodní den studentstva") \
            .on(month=11, day=17) \
            .since(2019) \
            .with_flags("NF")



from dateutil.easter import EASTER_ORTHODOX

from holidata.holiday import Country


class RU(Country):
    id = "RU"
    languages = ["ru"]
    default_lang = "ru"
    easter_type = EASTER_ORTHODOX

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Новый Год") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Рождество Христово") \
            .on(month=1, day=7) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("День защитника Отечества") \
            .on(month=2, day=23) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Международный женский день") \
            .on(month=3, day=8) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Праздник весны и труда") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("День Победы") \
            .on(month=5, day=9) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("День России") \
            .on(month=6, day=12) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("День народного единства") \
            .on(month=11, day=4) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Пасха") \
            .on(self.easter()) \
            .with_flags("NRV")



from holidata.holiday import Region
from holidata.utils import day


class EX(Region):
    def __init__(self, country):
        super().__init__("EX", country)

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on(month=1, day=7) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Martes de Carnaval") \
            .in_years([2023, 2024]) \
            .on(day(47).before(country.easter())) \
            .with_flags("V")

        self.define_holiday() \
            .with_name("Lunes siguiente a San José") \
            .in_years([2017]) \
            .on(month=3, day=20) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Día de Extremadura") \
            .except_for([2019, 2024]) \
            .on(month=9, day=8) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de Extremadura") \
            .in_years([2013, 2019]) \
            .on(month=9, day=9) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("San José") \
            .in_years([2021]) \
            .on(month=3, day=19) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta Nacional de España") \
            .in_years([2014]) \
            .on(month=10, day=13) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a Todos los Santos") \
            .in_years([2015, 2020]) \
            .on(month=11, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2015, 2020]) \
            .on(month=12, day=7) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a La Inmaculada Concepción") \
            .in_years([2013, 2019, 2024]) \
            .on(month=12, day=9) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Trabajo") \
            .in_years([2011, 2016, 2022]) \
            .on(month=5, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on(month=12, day=26) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2011, 2016]) \
            .on(month=12, day=26) \
            .with_flags("RF")



from dateutil.easter import EASTER_ORTHODOX

from holidata.holiday import Country
from holidata.utils import SmartDayArrow, day


class GR(Country):
    id = "GR"
    languages = ["el"]
    default_lang = "el"
    easter_type = EASTER_ORTHODOX

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Πρωτοχρονιά") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Θεοφάνεια") \
            .on(month=1, day=6) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Ευαγγελισμός της Θεοτόκου και Εθνική Ημέρα Ανεξαρτησίας της Ελλάδας") \
            .on(month=3, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Κοίμηση της Θεοτόκου") \
            .on(month=8, day=15) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Ημέρα του ΌΧΙ") \
            .on(month=10, day=28) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Χριστούγεννα") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Επόμενη ημέρα Χριστουγέννων") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Καθαρά Δευτέρα") \
            .on(day(48).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Μεγάλη Παρασκευή") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Μεγάλο Σάββατο") \
            .on(day(1).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Πάσχα") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Δευτέρα του Πάσχα") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Δευτέρα του Αγίου Πνεύματος") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Πρωτομαγιά") \
            .on(self.date_for_may_day) \
            .with_flags("NF")

    def date_for_may_day(self, year):
        """
        Postponed if it collides with Easter
        """
        date = SmartDayArrow(year, 5, 1)
        easter_date = self.easter()(year)

        if date == easter_date:
            date = date.shift(days=2)
        elif date == easter_date.shift(days=1):
            date = date.shift(days=1)

        return date



from holidata.holiday import Region
from holidata.utils import SmartDayArrow, day


class AN(Region):
    def __init__(self, country):
        super().__init__("AN", country)

        self.define_holiday() \
            .with_name("Lunes siguiente al Año Nuevo") \
            .in_years([2017, 2023]) \
            .on(month=1, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on(month=1, day=7) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Día de Andalucía") \
            .on(AN.dia_de_andalucia) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Asunción de la Virgen") \
            .in_years([2021]) \
            .on(month=8, day=16) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a La Inmaculada Concepción") \
            .in_years([2013, 2019, 2024]) \
            .on(month=12, day=9) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Trabajo") \
            .in_years([2011, 2016, 2022]) \
            .on(month=5, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2011, 2016]) \
            .on(month=12, day=26) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta Nacional de España") \
            .in_years([2014]) \
            .on(month=10, day=13) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a Todos los Santos") \
            .in_years([2015, 2020]) \
            .on(month=11, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2015, 2020]) \
            .on(month=12, day=7) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on(month=12, day=26) \
            .with_flags("RF")

    @staticmethod
    def dia_de_andalucia(year):
        if year == 2016:
            return SmartDayArrow(year, 2, 29)
        elif year == 2021:
            return SmartDayArrow(year, 3, 1)
        else:
            return SmartDayArrow(year, 2, 28)



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day

"""
sources
https://www.riigiteataja.ee/akt/109032011007 (Public Holidays and Days of National Importance Act)
"""


class EE(Country):
    id = "EE"
    languages = ["et"]
    default_lang = "et"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Uusaasta") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Iseseisvuspäev, Eesti Vabariigi aastapäev") \
            .on(month=2, day=24) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Kevadpüha") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Võidupüha") \
            .on(month=6, day=23) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Jaanipäev") \
            .on(month=6, day=24) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Taasiseseisvumispäev") \
            .on(month=8, day=20) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Jõululaupäev") \
            .on(month=12, day=24) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Esimene jõulupüha") \
            .on(month=12, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Teine jõulupüha") \
            .on(month=12, day=26) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Suur reede") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ülestõusmispühade 1. püha") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Nelipühade 1. püha") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")



import re

import pytest

from holidata import Locale, Country
from tests import HOLIDATA_YEAR_MAX


def create_locales():
    response = []

    for country_class in Country.plugins:
        response.extend([Locale(country_class(), lang_id) for lang_id in country_class.languages])

    return response


locales = create_locales()


@pytest.fixture(params=range(2011, HOLIDATA_YEAR_MAX))
def year(request):
    return request.param


@pytest.fixture(params=locales, ids=[loc.id for loc in locales])
def locale(request):
    return request.param


@pytest.fixture()
def holidays(locale, year):
    return locale.get_holidays_of(year)


@pytest.fixture(params=Country.plugins)
def country(request):
    return request.param()


@pytest.fixture()
def regions(locale):
    try:
        return [region.id for region in locale.country.regions]
    except AttributeError:
        return []


def test_country_should_be_constructable(country):
    pass


def test_holiday_should_not_be_of_type_national_if_region_defined(holidays):
    for holiday in holidays:
        if holiday.region == "":
            assert "N" in holiday.flags, f"Holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} must have flag 'N': it has no regions defined"


def test_holiday_should_be_of_type_national_if_no_region_defined(holidays):
    for holiday in holidays:
        if holiday.region != "":
            assert "N" not in holiday.flags, f"Holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} must not have flag 'N': it has regions defined"


def test_holiday_should_be_of_type_either_fixed_or_variable(holidays):
    for holiday in holidays:
        date_is_fixed = "F" in holiday.flags
        date_is_variable = "V" in holiday.flags

        assert not (date_is_variable and date_is_fixed), f"Holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} must not have both flags 'F' and 'V' (has '{holiday.flags}')"
        assert (date_is_variable or date_is_fixed), f"Holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} must have either flag 'F' or 'V' (has '{holiday.flags}')"


def test_holiday_flags_should_be_in_the_correct_order(holidays):
    for holiday in holidays:
        match = re.search(r"^N?R?[FV]?$", f"{holiday.flags}")

        assert match is not None, f"Flags for holiday '{holiday.description}' ({holiday.date.strftime('%Y-%m-%d')}) in locale {holiday.locale} are not in the correct order. Flags '{holiday.flags}' should match 'N?R?[FV]?'"


def test_holiday_should_only_be_defined_for_valid_regions(holidays, regions):
    if regions is None or regions == []:
        return

    for holiday in holidays:
        assert holiday.region == "" or holiday.region in regions, "Holiday '{}' ({}) in locale {} is defined for unknown region '{}'".format(holiday.description, holiday.date.strftime("%Y-%m-%d"), holiday.locale, holiday.region)



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class AT(Country):
    id = "AT"
    languages = ["de"]
    default_lang = "de"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Neujahr") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Heilige drei Könige") \
            .on(month=1, day=6) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Josef") \
            .in_regions(["2", "6", "7", "8"]) \
            .on(month=3, day=19) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Staatsfeiertag") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Florian") \
            .in_regions(["4"]) \
            .on(month=5, day=4) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Mariä Himmelfahrt") \
            .on(month=8, day=15) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Rupert") \
            .in_regions(["5"]) \
            .on(month=9, day=24) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Tag der Volksabstimmung") \
            .in_regions(["2"]) \
            .on(month=10, day=10) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Nationalfeiertag") \
            .on(month=10, day=26) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Allerheiligen") \
            .on(month=11, day=1) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Martin") \
            .in_regions(["1"]) \
            .on(month=11, day=11) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Leopold") \
            .in_regions(["9", "3"]) \
            .on(month=11, day=15) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Mariä Empfängnis") \
            .on(month=12, day=8) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Heiliger Abend") \
            .on(month=12, day=24) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Christtag") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Stefanitag") \
            .on(month=12, day=26) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Silvester") \
            .on(month=12, day=31) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Karfreitag") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ostersonntag") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ostermontag") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Christi Himmelfahrt") \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstsonntag") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstmontag") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Fronleichnam") \
            .on(day(60).after(self.easter())) \
            .with_flags("NRV")



from holidata.holiday import Region
from holidata.utils import day, fourth


class SN(Region):
    def __init__(self, country):
        super().__init__("SN", country)

        """
        11 days before 4. sunday before 12-25: Buß- und Bettag
        """
        self.define_holiday() \
            .with_name("Buß- und Bettag") \
            .on(day(11).before(fourth("sunday").before(month=12, day=25))) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Reformationstag") \
            .until(2016) \
            .on(month=10, day=31) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Reformationstag") \
            .since(2018) \
            .on(month=10, day=31) \
            .with_flags("RF")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day, first


class CO(Country):
    id = "CO"
    languages = ["es"]
    default_lang = "es"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Año Nuevo") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Día del Trabajo") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Grito de Independencia") \
            .on(month=7, day=20) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Batalla de Boyacá") \
            .on(month=8, day=7) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Inmaculada Concepción") \
            .on(month=12, day=8) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Navidad") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Viernes Santo") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Domingo de Pascua") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("La Ascensión del Señor") \
            .on(day(43).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Corpus Christi") \
            .on(day(64).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("El Sagrado Corazón de Jesús") \
            .on(day(71).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Día de los Reyes Magos") \
            .on(first("monday").after(month=1, day=6, including=True)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Día de San José") \
            .on(first("monday").after(month=3, day=19, including=True)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("San Pedro y San Pablo") \
            .on(first("monday").after(month=6, day=29, including=True)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Asunción de la Virgen") \
            .on(first("monday").after(month=8, day=15, including=True)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Día de la Raza") \
            .on(first("monday").after(month=10, day=12, including=True)) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Todos los Santos") \
            .on(first("monday").after(month=11, day=1, including=True)) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Independencia de Cartagena") \
            .on(first("monday").after(month=11, day=11, including=True)) \
            .with_flags("NV")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class PT(Country):
    id = "PT"
    languages = ["pt"]
    default_lang = "pt"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Ano Novo") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dia da Liberdade") \
            .on(month=4, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dia do Trabalhador") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dia de Portugal") \
            .on(month=6, day=10) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Assunção de Nossa Senhora") \
            .on(month=8, day=15) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Implantação da República") \
            .on(month=10, day=5) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dia de Todos os Santos") \
            .on(month=11, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Restauração da Independência") \
            .on(month=12, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Imaculada Conceição") \
            .on(month=12, day=8) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Natal") \
            .on(month=12, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Carnaval") \
            .on(day(47).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Sexta-feira Santa") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Páscoa") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Corpo de Deus") \
            .on(day(60).after(self.easter())) \
            .with_flags("NRV")



"""
Provides date-handling related utils.
"""
from typing import Callable

from arrow import Arrow

WEEKDAYS = [
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday"
]

MONTHS = [
    "january",
    "february",
    "march",
    "april",
    "may",
    "june",
    "july",
    "august",
    "september",
    "october",
    "november",
    "december"
]


class SmartDayArrow(Arrow):
    """
    A wrapper around Arrow datetime reference that provides additional convenience methods.
    """

    def weekday(self) -> str:
        """
        Provide a more readable weekday representation.
        """

        return WEEKDAYS[super().weekday()]

    def shift_to_weekday(self, day: str, order: int = 1, reverse: bool = False, including: bool = False) -> 'SmartDayArrow':
        """
        Shifts to {order}. weekday in the given direction, i.e. 2. monday before this date would be:

        >>> arrow.shift_to_weekday("monday", order=2, reverse=True)
        """

        result = self

        if including and result.weekday() == day:
            if order == 1:
                return result
            else:
                order = order - 1

        while order > 0:
            result = result.shift(days=1 if not reverse else -1)
            if day == result.weekday():
                order = order - 1

        return result


def date(month: int, day: int) -> Callable[[int], SmartDayArrow]:
    def wrapper(year: int):
        return SmartDayArrow(year, month, day)

    return wrapper


class SmartDayArrowDayShifter:
    def __init__(self, days):
        self.day_count: int = days
        self.shift_direction: int = 0
        self.date: callable = None

    def before(self, date: callable) -> 'SmartDayArrowDayShifter':
        self.date = date
        self.shift_direction = -1
        return self

    def after(self, date: callable) -> 'SmartDayArrowDayShifter':
        self.date = date
        self.shift_direction = 1
        return self

    def __call__(self, year: int) -> SmartDayArrow:
        return self.date(year).shift(days=self.day_count * self.shift_direction)


def day(count: int) -> SmartDayArrowDayShifter:
    return SmartDayArrowDayShifter(count)


class SmartDayArrowWeekdayShifter:
    def __init__(self, weekday: str, order: int, forward: bool):
        self.weekday = weekday
        self.order = order
        self.forward = forward
        self.including = True
        self.date = None

    def of(self, month: str) -> 'SmartDayArrowWeekdayShifter':
        month_index = MONTHS.index(month.lower()) + 1

        if self.forward:
            self.date = date(month_index, 1)
        else:
            def wrapper(year):
                return SmartDayArrow(
                    year if month_index != 12 else year + 1,
                    (month_index % 12) + 1,
                    1).shift(days=-1)

            self.date = wrapper

        return self

    def before(self, date: callable = None, month: int = None, day: int = None, including: bool = False) -> 'SmartDayArrowWeekdayShifter':
        self._configure_shift(date, month, day, False, including)
        return self

    def after(self, date: callable = None, month: int = None, day: int = None, including: bool = False) -> 'SmartDayArrowWeekdayShifter':
        self._configure_shift(date, month, day, True, including)
        return self

    def _configure_shift(self, date_func: callable, month: int, day: int, forward: bool, including: bool):
        if month is not None and day is not None:
            self.date = date(month, day)
        elif callable(date_func):
            self.date = date_func
        else:
            raise ValueError("Invalid reference date")

        self.forward = forward
        self.including = including

    def __call__(self, year: int) -> SmartDayArrow:
        return self.date(year).shift_to_weekday(
            self.weekday,
            order=self.order,
            reverse=not self.forward,
            including=self.including,
        )


def first(weekday: str) -> SmartDayArrowWeekdayShifter:
    return SmartDayArrowWeekdayShifter(weekday, order=1, forward=True)


def second(weekday: str) -> SmartDayArrowWeekdayShifter:
    return SmartDayArrowWeekdayShifter(weekday, order=2, forward=True)


def third(weekday: str) -> SmartDayArrowWeekdayShifter:
    return SmartDayArrowWeekdayShifter(weekday, order=3, forward=True)


def fourth(weekday: str) -> SmartDayArrowWeekdayShifter:
    return SmartDayArrowWeekdayShifter(weekday, order=4, forward=True)


def last(weekday: str) -> SmartDayArrowWeekdayShifter:
    return SmartDayArrowWeekdayShifter(weekday, order=1, forward=False)



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day, first, third, second


#
# Information taken from various government websites around 2020-04-16
#   https://www.canada.ca/fr/emploi-developpement-social/services/normes-travail/rapports/jours-feries.html
#   http://www4.gouv.qc.ca/FR/Portail/Citoyens/Evenements/travailleur-avec-salaire/Pages/jours-feries-chomes-payes.aspx
#   https://www.ontario.ca/document/your-guide-employment-standards-act-0/public-holidays#section-2
#   https://www.cfib-fcei.ca/en/tools-resources/paying-employees-public-holidays-newfoundland-labrador
#   https://www.princeedwardisland.ca/en/information/economic-growth-tourism-and-culture/paid-holidays
#   https://www2.gnb.ca/content/dam/gnb/Departments/petl-epft/PDF/es/FactSheets/PublicHolidaysVacation.pdf
#   https://www.gov.mb.ca/labour/standards/doc,gen-holidays-after-april-30-07,factsheet.html#q12
#   https://www.saskatchewan.ca/business/employment-standards/vacations-holidays-leaves-and-absences/public-statutory-holidays/list-of-saskatchewan-public-holidays
#
# Also those sites for some information
#   https://www.officeholidays.com/holidays/canada/canada-remembrance-day
#   https://www.timeanddate.com/holidays/canada/family-day
#   https://fr.wikipedia.org/wiki/F%C3%AAtes_et_jours_f%C3%A9ri%C3%A9s_au_Canada
#
# I have not checked every province and territory website, but the wikipedia
# summary has been true for everything that I have checked, although it seems
# to be considering more holidays than the bare minimum and also counts holidays
# that are not mandated but usually observed (e.g. St-Patrick's Day in NL which
# is not statutory, but is given to government employees).
#


class CA(Country):
    id = "CA"
    languages = ["en", "fr"]
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_names({
                "en": "New Year's Day",
                "fr": "Jour de l'An",
            }) \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "en": "National Holiday",
                "fr": "Fête Nationale",
            }) \
            .in_regions(["QC"]) \
            .on(month=6, day=24) \
            .with_flags("F")

        self.define_holiday() \
            .with_names({
                "en": "Canada Day",
                "fr": "Fête du Canada",
            }) \
            .on(month=7, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_names({
                "en": "Remembrance Day",
                "fr": "Jour du Souvenir",
            }) \
            .in_regions(["AB", "BC", "NB", "NL", "NT"]) \
            .on(month=11, day=11) \
            .with_flags("F")

        self.define_holiday() \
            .with_names({
                "en": "Christmas Day",
                "fr": "Jour de Noël",
            }) \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "en": "Boxing Day",
                "fr": "Lendemain de Noël",
            }) \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_names({
                "en": "Good Friday",
                "fr": "Vendredi Saint",
            }) \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_names({
                "en": "Easter Monday",
                "fr": "Lundi de Pâques",
            }) \
            .in_regions(["AB", "PE", "QC"]) \
            .on(day(1).after(self.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_names({
                "en": "Family Day",
                "fr": "Fête de la Famille",
            }) \
            .in_regions(["AB", "ON", "SK", "NB"]) \
            .on(third("monday").of("february")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "Louis Riel Day",
                "fr": "Journée Louis Riel",
            }) \
            .in_regions(["MB"]) \
            .on(third("monday").of("february")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "Islander Day",
                "fr": "Fête des Insulaires",
            }) \
            .in_regions(["PE"]) \
            .on(third("monday").of("february")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "August Civic Holiday",
                "fr": "Premier lundi d'août",
            }) \
            .in_regions(["NT", "NU"]) \
            .on(first("monday").of("august")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "Saskatchewan Day",
                "fr": "Fête de la Saskatchewan",
            }) \
            .in_regions(["SK"]) \
            .on(first("monday").of("august")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "Heritage Day",
                "fr": "Fête du Patrimoine",
            }) \
            .in_regions(["AB"]) \
            .on(first("monday").of("august")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "Heritage Day",
                "fr": "Jour de la Fondation",
            }) \
            .in_regions(["NS"]) \
            .on(first("monday").of("august")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "New Brunswick Day",
                "fr": "Jour du Nouveau-Brunswick",
            }) \
            .in_regions(["NB"]) \
            .on(first("monday").of("august")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "Labour Day",
                "fr": "Fête du Travail",
            }) \
            .on(first("monday").of("september")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_names({
                "en": "Thanksgiving Day",
                "fr": "Jour de l'Action de grâce",
            }) \
            .in_regions(["AB", "BC", "MB", "NL", "ON", "QC", "SK", "NT", "NU", "YT"]) \
            .on(second("monday").of("october")) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "National Patriots' Day",
                "fr": "Journée Nationale des Patriotes",
            }) \
            .in_regions(["QC"]) \
            .on(first("monday").before(month=5, day=25)) \
            .with_flags("V")

        self.define_holiday() \
            .with_names({
                "en": "Victoria Day",
                "fr": "Fête de la Reine Victoria",
            }) \
            .in_regions(["AB", "BC", "MB", "NS", "ON", "SK", "NT", "NU", "YT"]) \
            .on(first("monday").before(month=5, day=25)) \
            .with_flags("V")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day, first


class IS(Country):
    id = "IS"
    languages = ["is"]
    default_lang = "is"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nýársdagur") \
            .on(month=1, day=1) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Verkalýðsdagurinn") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Þjóðhátíðardagurinn") \
            .on(month=6, day=17) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Jóladagur") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Annar dagur jóla") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Skírdagur") \
            .on(day(3).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Föstudagurinn langi") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Páskadagur") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Annar dagur páska") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Uppstigningardagur") \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Hvítasunnudagur") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Annar dagur hvítasunnu") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Frídagur verslunarmanna") \
            .on(first("monday").of("august")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Sumardagurinn fyrsti") \
            .on(first("thursday").after(month=4, day=18)) \
            .with_flags("NV")

        """ 
        Both, Christmas Eve (_aðfangadagur jóla_) and New Year's Eve (_gamlársdagur_) are public holidays in Iceland from 13:00 only.
        They're included as full-day holidays, but with an explanatory note.
        """
        self.define_holiday() \
            .with_name("Aðfangadagur jóla") \
            .on(month=12, day=24) \
            .with_flags("NRF") \
            .annotated_with("Holiday from 13:00")

        self.define_holiday() \
            .with_name("Gamlársdagur") \
            .on(month=12, day=31) \
            .with_flags("NF") \
            .annotated_with("Holiday from 13:00")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import SmartDayArrow, day


class NL(Country):
    id = "NL"
    languages = ["nl"]
    default_lang = "nl"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nieuwjaarsdag") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dodenherdenking") \
            .on(month=5, day=4) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Bevrijdingsdag") \
            .on(month=5, day=5) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Sinterklaas") \
            .on(month=12, day=5) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Eerste Kerstdag") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Tweede Kerstdag") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Goede Vrijdag") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Eerste Paasdag") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Tweede Paasdag") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Hemelvaartsdag") \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Eerste Pinksterdag") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Tweede Pinksterdag") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Koninginnedag") \
            .until(2013) \
            .on(NL.koninginnedag) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Koningsdag") \
            .since(2014) \
            .on(NL.koningsdag) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Koninkrijksdag") \
            .on(NL.koninkrijksdag) \
            .with_flags("NV")

    @staticmethod
    def koninginnedag(year):
        """
        04-30 or saturday before, if it falls on sunday
        """
        date = SmartDayArrow(year, 4, 30)

        if date.weekday() == "sunday":
            date = date.shift(days=-1)

        return date

    @staticmethod
    def koningsdag(year):
        """
        04-27 or saturday before if it falls on sunday
        """
        date = SmartDayArrow(year, 4, 27)

        if date.weekday() == "sunday":
            date = date.shift(days=-1)

        return date

    @staticmethod
    def koninkrijksdag(year):
        """
        04-27 or monday after if it falls on sunday
        """
        date = SmartDayArrow(year, 12, 15)

        if date.weekday() == "sunday":
            date = date.shift(days=1)

        return date



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class FR(Country):
    id = "FR"
    languages = ["fr"]
    default_lang = "fr"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Jour de l'an") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Fête du premier mai") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Armistice 1945") \
            .on(month=5, day=8) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Fête nationale") \
            .on(month=7, day=14) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Assomption") \
            .on(month=8, day=15) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Toussaint") \
            .on(month=11, day=1) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Armistice 1918") \
            .on(month=11, day=11) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Noël") \
            .on(month=12, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Lundi de Pâques") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ascension") \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pentecôte") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Lundi de Pentecôte") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")



import pytest
from syrupy import snapshot
from syrupy.extensions.json import JSONSnapshotExtension

from holidata import Locale, Country
from tests import HOLIDATA_YEAR_MAX


def create_locales():
    response = []

    for country_class in Country.plugins:
        response.extend([Locale(country_class(), lang_id) for lang_id in country_class.languages])

    return response


locales = create_locales()


@pytest.fixture
def snapshot_json(snapshot):
    return snapshot.with_defaults(extension_class=JSONSnapshotExtension)


@pytest.fixture(params=range(2011, HOLIDATA_YEAR_MAX))
def year(request):
    return request.param


@pytest.fixture(params=locales, ids=[loc.id for loc in locales])
def locale(request):
    return request.param


def test_holidata_produces_holidays_for_locale_and_year(snapshot_json, locale, year):
    export_data = [h.as_dict() for h in locale.get_holidays_of(year)]
    export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))

    assert export_data == snapshot_json


def test_holidata_can_produce_holidays_out_of_range(locale):
    [h.as_dict() for h in locale.get_holidays_of(HOLIDATA_YEAR_MAX)]



from holidata.holiday import Region
from holidata.utils import day


class NC(Region):
    def __init__(self, country):
        super().__init__("NC", country)

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on(month=1, day=7) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San José") \
            .in_years([2012, 2014, 2015, 2019, 2020, 2021]) \
            .on(month=3, day=19) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Santiago Apóstol / Día Nacional de Galicia") \
            .in_years([2023, 2024]) \
            .on(month=7, day=25) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes de Pascua") \
            .on(day(1).after(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Santiago Apóstol") \
            .in_years([2011, 2013, 2015, 2016, 2017, 2022]) \
            .on(month=7, day=25) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2020]) \
            .on(month=12, day=7) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on(month=12, day=26) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2011, 2016]) \
            .on(month=12, day=26) \
            .with_flags("RF")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.holidays.DE.BB import BB
from holidata.holidays.DE.BE import BE
from holidata.holidays.DE.BW import BW
from holidata.holidays.DE.BY import BY
from holidata.holidays.DE.HB import HB
from holidata.holidays.DE.HE import HE
from holidata.holidays.DE.HH import HH
from holidata.holidays.DE.MV import MV
from holidata.holidays.DE.NI import NI
from holidata.holidays.DE.NW import NW
from holidata.holidays.DE.RP import RP
from holidata.holidays.DE.SH import SH
from holidata.holidays.DE.SL import SL
from holidata.holidays.DE.SN import SN
from holidata.holidays.DE.ST import ST
from holidata.holidays.DE.TH import TH
from holidata.utils import day

__all__ = [
    "DE",
]


class DE(Country):
    id = "DE"
    languages = ["de"]
    default_lang = "de"
    easter_type = EASTER_WESTERN

    """
    https://www.bmi.bund.de/DE/themen/verfassung/staatliche-symbole/nationale-feiertage/nationale-feiertage-node.html
    """
    def __init__(self):
        super().__init__()

        self.regions = [
            BB(self),
            BE(self),
            BW(self),
            BY(self),
            HB(self),
            HE(self),
            HH(self),
            MV(self),
            NI(self),
            NW(self),
            RP(self),
            SH(self),
            SL(self),
            SN(self),
            ST(self),
            TH(self),
        ]

        self.define_holiday() \
            .with_name("Neujahr") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Erster Maifeiertag") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Tag der Deutschen Einheit") \
            .on(month=10, day=3) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Heilig Abend") \
            .on(month=12, day=24) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Weihnachtstag") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Zweiter Weihnachtstag") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Silvester") \
            .on(month=12, day=31) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Karfreitag") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ostern") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Ostermontag") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Christi Himmelfahrt") \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstsonntag") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pfingstmontag") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")

        """
        2017-10-31: Reformationstag (national holiday because of 500th anniversary)
        """
        self.define_holiday() \
            .with_name("Reformationstag") \
            .in_years([2017]) \
            .on(month=10, day=31) \
            .with_flags("NRF")



from holidata.holiday import Region
from holidata.utils import SmartDayArrow, day


class CE(Region):
    def __init__(self, country):
        super().__init__("CE", country)

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on(month=1, day=7) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Nuestra Señora de África") \
            .in_years([2022, 2023, 2024]) \
            .on(month=8, day=5) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Día de Ceuta") \
            .in_years([2016, 2017, 2019, 2020, 2021, 2022, 2023]) \
            .on(month=9, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Fiesta del Sacrificio (Eidul Adha)") \
            .except_for([2011]) \
            .on(CE.day_of_eidul_adha) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Sacrificio (Eidul Adha)") \
            .in_years([2011]) \
            .on(month=11, day=7) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Trabajo") \
            .in_years([2011]) \
            .on(month=5, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta Nacional de España") \
            .in_years([2014]) \
            .on(month=10, day=13) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a Todos los Santos") \
            .in_years([2015]) \
            .on(month=11, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2015, 2020]) \
            .on(month=12, day=7) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a La Inmaculada Concepción") \
            .in_years([2013]) \
            .on(month=12, day=9) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2011, 2016]) \
            .on(month=12, day=26) \
            .with_flags("RF")

    @staticmethod
    def day_of_eidul_adha(year):
        dates = {
            2012: SmartDayArrow(year, 10, 27),
            2013: SmartDayArrow(year, 10, 15),
            2014: SmartDayArrow(year, 10, 6),
            2015: SmartDayArrow(year, 9, 25),
            2016: SmartDayArrow(year, 9, 12),
            2017: SmartDayArrow(year, 9, 1),
            2018: SmartDayArrow(year, 8, 22),
            2019: SmartDayArrow(year, 8, 12),
            2020: SmartDayArrow(year, 7, 31),
            2021: SmartDayArrow(year, 7, 20),
            2022: SmartDayArrow(year, 7, 9),
            2023: SmartDayArrow(year, 6, 29),
            2024: SmartDayArrow(year, 6, 17),
        }

        return dates.get(year)



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class NO(Country):
    id = "NO"
    languages = ["nb"]
    default_lang = "nb"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nyttårsdag") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Offentlig Høytidsdag") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Frigjøringsdag 1945") \
            .on(month=5, day=8) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Grunnlovsdag") \
            .on(month=5, day=17) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Julaften") \
            .on(month=12, day=24) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Juledag") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Juledag") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Nyttårsaften") \
            .on(month=12, day=31) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Fastelavn") \
            .on(day(49).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Palmesøndag") \
            .on(day(7).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Skjærtorsdag") \
            .on(day(3).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Langfredag") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Påskedag") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Påskedag") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Kristi Himmelfartsdag") \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pinsedag") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pinsedag") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


# Information gathered from Brazilian legislation.

# National holidays:
# Law 10.607/2002: http://www.planalto.gov.br/ccivil_03/leis/2002/l10607.htm
# Law 6.802/1980:  http://www.planalto.gov.br/ccivil_03/leis/l6802.htm

# State holidays
# AC: State laws 1.538/2004, 1.411/2001, 14/1964, 1.526/2004, and 57/1965
# AL: State laws 5.508/1993, 5.509/1993, 5.724/1995, and Decree 68.782 (30/dec/2019)
# AP: State law 667/2002 and Art. 355 of the Constitution of the Federative Republic of Brazil
# AM: State laws 25/1977 and 84/2010
# BA: Art. 6 of the Constitution of the Federative Republic of Brazil
# CE: State laws 9.093/1995, 9.093/1995 and Art. 18 of the Constitution of the State of Ceará
# DF: District law 963/1995
# ES: State law 11.010/2019
# GO: State law 20.756/2020
# MA: State law 2.457/1964
# MT: State law 7.879/2002
# MS: State law 10/1979
# MG: Art. 256 of the Constitution of the State of Minas Gerais
# PA: State law 5.999/1996
# PB: State law 3.489/1967
# PR: State law 4.658/1962
# PE: State law 13.835/2009
# PI: State law 176/1937
# RJ: State laws 5.243/2008, 5.198/2008, and 4.007/2002
# RN: State laws 7.831/2000 and 8.913/2006
# RS: Art. 6 of the Constitution of the State of Rio Grande do Sul and Decree 36.180 (18/set/1995)
# RO: State laws 2.291/2010 and 1.026/2001
# RR: Art. 9 of the Constitution of the State of Roraima
# SC: State laws 12.906/2004 and 10.306/1996
# SP: State law 9.497/1997
# SE: Art. 269 of the Constitution of the State of Sergipe
# TO: State laws 98/1989, 960/1998, 627/1993


class BR(Country):
    id = "BR"
    languages = ["pt"]
    default_lang = "pt"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()
        
        self.define_holiday() \
            .with_name("Confraternização Universal") \
            .on(month=1, day=1) \
            .with_flags("NF")
        
        self.define_holiday() \
            .with_name("Criação do Estado de Rondônia") \
            .in_regions(["RO"]) \
            .on(month=1, day=4) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia do Evangélico no Acre") \
            .in_regions(["AC"]) \
            .on(month=1, day=23) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("Revolução Pernambucana de 1817") \
            .in_regions(["PE"]) \
            .on(month=3, day=6) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia Internacional da Mulher") \
            .in_regions(["AC"]) \
            .on(month=3, day=8) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Autonomia do Estado de Tocantins") \
            .in_regions(["TO"]) \
            .on(month=3, day=18) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia de São José") \
            .in_regions(["AP", "CE"]) \
            .on(month=3, day=19) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("Abolição da Escravidão no Ceará") \
            .in_regions(["CE"]) \
            .on(month=3, day=25) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Fundação de Brasília") \
            .in_regions(["DF"]) \
            .on(month=4, day=21) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Execução de Tiradentes") \
            .in_regions(["MG"]) \
            .on(month=4, day=21) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Tiradentes") \
            .on(month=4, day=21) \
            .with_flags("NF")
        
        self.define_holiday() \
            .with_name("Dia de São Jorge") \
            .in_regions(["RJ"]) \
            .on(month=4, day=23) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("Dia Internacional do Trabalhador") \
            .on(month=5, day=1) \
            .with_flags("NF")
        
        self.define_holiday() \
            .with_name("Aniversário do Estado do Acre") \
            .in_regions(["AC"]) \
            .on(month=6, day=15) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia do Evangélico em Rondônia") \
            .in_regions(["RO"]) \
            .on(month=6, day=18) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("São João") \
            .in_regions(["AL", "PE"]) \
            .on(month=6, day=24) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("São Pedro") \
            .in_regions(["AL"]) \
            .on(month=6, day=29) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("Independência da Bahia") \
            .in_regions(["BA"]) \
            .on(month=7, day=2) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Emancipação Política de Sergipe") \
            .in_regions(["SE"]) \
            .on(month=7, day=8) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Revolução Constitucionalista de 1932") \
            .in_regions(["SP"]) \
            .on(month=7, day=9) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Fundação da Cidade de Goiás") \
            .in_regions(["GO"]) \
            .on(month=7, day=26) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Adesão do Maranhão à Independência do Brasil") \
            .in_regions(["MA"]) \
            .on(month=7, day=28) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Fundação do Estado da Paraíba") \
            .in_regions(["PB"]) \
            .on(month=8, day=5) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia do Rio Grande do Norte") \
            .in_regions(["RN"]) \
            .on(month=8, day=7) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia de Santa Catarina") \
            .in_regions(["SC"]) \
            .on(month=8, day=11) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia de Nossa Senhora da Assunção") \
            .in_regions(["CE"]) \
            .on(month=8, day=15) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("Adesão do Pará à Independência do Brasil") \
            .in_regions(["PA"]) \
            .on(month=8, day=15) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia da Amazônia") \
            .in_regions(["AC"]) \
            .on(month=9, day=5) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Elevação do Amazonas à Categoria de Província") \
            .in_regions(["AM"]) \
            .on(month=9, day=5) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Independência do Brasil") \
            .on(month=9, day=7) \
            .with_flags("NF")
        
        self.define_holiday() \
            .with_name("Nossa Senhora da Natividade") \
            .in_regions(["TO"]) \
            .on(month=9, day=8) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Criação do Território Federal do Amapá") \
            .in_regions(["AP"]) \
            .on(month=9, day=13) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Emancipação Política do Alagoas") \
            .in_regions(["AL"]) \
            .on(month=9, day=16) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia do Gaúcho") \
            .in_regions(["RS"]) \
            .on(month=9, day=20) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Mártires de Cunhaú e Uruaçu") \
            .in_regions(["RN"]) \
            .on(month=10, day=3) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Criação dos Estado de Roraima") \
            .in_regions(["RR"]) \
            .on(month=10, day=5) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Criação dos Estado de Tocantins") \
            .in_regions(["TO"]) \
            .on(month=10, day=5) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Criação do Estado do Mato Grosso do Sul") \
            .in_regions(["MS"]) \
            .on(month=10, day=11) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Nossa Senhora Aparecida") \
            .on(month=10, day=12) \
            .with_flags("NRF")
        
        self.define_holiday() \
            .with_name("Dia do Piauí") \
            .in_regions(["PI"]) \
            .on(month=10, day=19) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Pedra Fundamental de Goiânia") \
            .in_regions(["GO"]) \
            .on(month=10, day=24) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Finados") \
            .on(month=11, day=2) \
            .with_flags("NRF")
        
        self.define_holiday() \
            .with_name("Proclamação da República") \
            .on(month=11, day=15) \
            .with_flags("NF")
        
        self.define_holiday() \
            .with_name("Assinatura do Tratado de Petrópolis") \
            .in_regions(["AC"]) \
            .on(month=11, day=17) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Morte de Zumbi dos Palmares") \
            .in_regions(["AL"]) \
            .on(month=11, day=20) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia da Consciência Negra") \
            .in_regions(["AM", "MT", "RJ"]) \
            .on(month=11, day=20) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Dia de Santa Catarina de Alexandria") \
            .in_regions(["SC"]) \
            .on(month=11, day=25) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("Dia do Evangélico do Distrito Federal") \
            .in_regions(["DF"]) \
            .on(month=11, day=30) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("Nossa Senhora da Conceição") \
            .in_regions(["AM"]) \
            .on(month=12, day=8) \
            .with_flags("RF")
        
        self.define_holiday() \
            .with_name("Emancipação Política do Estado do Paraná") \
            .in_regions(["PR"]) \
            .on(month=12, day=19) \
            .with_flags("F")
        
        self.define_holiday() \
            .with_name("Natal") \
            .on(month=12, day=25) \
            .with_flags("NRF")
        
        self.define_holiday() \
            .with_name("Carnaval") \
            .on(day(47).before(self.easter())) \
            .with_flags("NRV")
        
        self.define_holiday() \
            .with_name("Páscoa") \
            .on(self.easter()) \
            .with_flags("NRV")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class IT(Country):
    id = "IT"
    languages = ["it"]
    default_lang = "it"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Capodanno") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Epifania") \
            .on(month=1, day=6) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Festa della liberazione") \
            .on(month=4, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Festa del lavoro") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Festa della repubblica") \
            .on(month=6, day=2) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Assunzione (ferragosto)") \
            .on(month=8, day=15) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Ognissanti") \
            .on(month=11, day=1) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Immacolata concezione") \
            .on(month=12, day=8) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Natale") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("S.to Stefano") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Pasqua") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pasquetta") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")



from holidata.holiday import Region
from holidata.utils import SmartDayArrow, day


class ML(Region):
    def __init__(self, country):
        super().__init__("ML", country)

        self.define_holiday() \
            .with_name("Lunes siguiente al Año Nuevo") \
            .in_years([2017]) \
            .on(month=1, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on(month=1, day=7) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Estatuto de Autonomía de la Ciudad de Melilla") \
            .in_years([2020, 2021]) \
            .on(month=3, day=13) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Fiesta del Eid Fitr") \
            .in_years([2022, 2023]) \
            .on(ML.day_of_eid_fitr) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Fiesta del Sacrificio (Aid Al Adha)") \
            .in_years([2022, 2023, 2024]) \
            .on(ML.holiday_aid_al_adha) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Fiesta del Sacrificio (Aid El Kebir)") \
            .until(2021) \
            .on(ML.holiday_aid_el_kebir) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("San José") \
            .in_years([2011, 2012, 2013, 2014, 2015, 2016]) \
            .on(month=3, day=19) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2015, 2020]) \
            .on(month=12, day=7) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a La Inmaculada Concepción") \
            .in_years([2019, 2024]) \
            .on(month=12, day=9) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on(month=12, day=26) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2011, 2016]) \
            .on(month=12, day=26) \
            .with_flags("RF")

    @staticmethod
    def holiday_aid_el_kebir(year):
        dates = {
            2011: SmartDayArrow(year, 11, 7),
            2012: SmartDayArrow(year, 10, 26),
            2013: SmartDayArrow(year, 10, 15),
            2014: SmartDayArrow(year, 10, 4),
            2015: SmartDayArrow(year, 9, 25),
            2016: SmartDayArrow(year, 9, 12),
            2017: SmartDayArrow(year, 9, 1),
            2018: SmartDayArrow(year, 8, 22),
            2019: SmartDayArrow(year, 8, 12),
            2020: SmartDayArrow(year, 7, 31),
            2021: SmartDayArrow(year, 7, 21),
        }
        return dates.get(year)

    @staticmethod
    def holiday_aid_al_adha(year):
        dates = {
            2022: SmartDayArrow(year, 7, 11),
            2023: SmartDayArrow(year, 6, 29),
            2024: SmartDayArrow(year, 6, 17),
        }
        return dates.get(year)

    @staticmethod
    def day_of_eid_fitr(year):
        dates = {
            2022: SmartDayArrow(year, 5, 3),
            2023: SmartDayArrow(year, 4, 21),
        }
        return dates.get(year)



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class HR(Country):
    id = "HR"
    languages = ["hr"]
    default_lang = "hr"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nova Godina") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Sveta tri kralja") \
            .on(month=1, day=6) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Praznik rada") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan državnosti") \
            .on(month=5, day=30) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan antifašističke borbe") \
            .on(month=6, day=22) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan pobjede i domovinske zahvalnosti i Dan hrvatskih branitelja") \
            .on(month=8, day=5) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Velika Gospa") \
            .on(month=8, day=15) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Dan svih svetih") \
            .on(month=11, day=1) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Božić") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Sveti Stjepan") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Uskrs") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Uskršnji ponedjeljak") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Tijelovo") \
            .on(day(60).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Dan sjećanja na žrtve Domovinskog rata i Dan sjećanja na žrtvu Vukovara i Škabrnje") \
            .since(2020) \
            .on(month=11, day=18) \
            .with_flags("NF")



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class SI(Country):
    id = "SI"
    languages = ["sl"]
    default_lang = "sl"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Novo leto") \
            .on(month=1, day=1) \
            .with_flags("NF")

        """
        From 1955 until May 2012, when the National Assembly of Slovenia passed the Public Finance Balance Act,
        2 January was a work-free day. It was reintroduced in 2017.
        2012<: https://www.uradni-list.si/1/objava.jsp?sop=2012-01-1700
        2016<: https://www.uradni-list.si/1/objava.jsp?sop=2016-01-3568
        """
        self.define_holiday() \
            .with_name("Novo leto") \
            .on(month=1, day=2) \
            .except_for([2013, 2014, 2015, 2016]) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Prešernov dan") \
            .on(month=2, day=8) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan upora proti okupatorju") \
            .on(month=4, day=27) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Praznik dela") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Praznik dela") \
            .on(month=5, day=2) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan državnosti") \
            .on(month=6, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Marijino vnebovzetje") \
            .on(month=8, day=15) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Dan reformacije") \
            .on(month=10, day=31) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Dan spomina na mrtve") \
            .on(month=11, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Božič") \
            .on(month=12, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Dan samostojnosti in enotnosti") \
            .on(month=12, day=26) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Velikonočna nedelja") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Velikonočni ponedeljek") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Binkošti") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")



from holidata.holiday import Region
from holidata.utils import SmartDayArrow, day


class AR(Region):
    def __init__(self, country):
        super().__init__("AR", country)

        self.define_holiday() \
            .with_name("Lunes siguiente al Año Nuevo") \
            .in_years([2017, 2023]) \
            .on(month=1, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on(month=1, day=7) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Jorge / Día de Aragón") \
            .on(AR.holiday_san_jorge__dia_de_aragon) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a San Jorge / Día de Aragón") \
            .in_years([2023]) \
            .on(month=4, day=24) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Asunción de la Virgen") \
            .in_years([2021]) \
            .on(month=8, day=16) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a La Inmaculada Concepción") \
            .in_years([2013, 2019, 2024]) \
            .on(month=12, day=9) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Trabajo") \
            .in_years([2011, 2016, 2022]) \
            .on(month=5, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta Nacional de España") \
            .in_years([2014]) \
            .on(month=10, day=13) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a Todos los Santos") \
            .in_years([2015, 2020]) \
            .on(month=11, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2015, 2020]) \
            .on(month=12, day=7) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on(month=12, day=26) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2011, 2016]) \
            .on(month=12, day=26) \
            .with_flags("RF")

    @staticmethod
    def holiday_san_jorge__dia_de_aragon(year):
        if year == 2017:
            return SmartDayArrow(year, 4, 24)
        else:
            return SmartDayArrow(year, 4, 23)



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import SmartDayArrow, day, first, last


class GB(Country):
    id = "GB"
    languages = ["en"]
    default_lang = "en"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("New Year's Day") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("New Year's Day (observed)") \
            .on(first("monday").after(month=1, day=1)) \
            .with_flags("NV") \
            .on_condition(GB.date_is_on_weekend(month=1, day=1))

        self.define_holiday() \
            .with_name("Christmas Day") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Christmas Day (observed)") \
            .on(first("monday").after(month=12, day=25)) \
            .with_flags("NV")  \
            .on_condition(GB.date_is_on_saturday(month=12, day=25))

        self.define_holiday() \
            .with_name("Christmas Day (observed)") \
            .on(first("tuesday").after(month=12, day=25)) \
            .with_flags("NV")  \
            .on_condition(GB.date_is_on_sunday(month=12, day=25))

        self.define_holiday() \
            .with_name("Boxing Day") \
            .on(month=12, day=26) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Boxing Day (observed)") \
            .on(first("monday").after(month=12, day=26)) \
            .with_flags("NV")  \
            .on_condition(self.date_is_on_saturday(month=12, day=26))

        self.define_holiday() \
            .with_name("Boxing Day (observed)") \
            .on(first("tuesday").after(month=12, day=26)) \
            .with_flags("NV")  \
            .on_condition(GB.date_is_on_sunday(month=12, day=26))

        self.define_holiday() \
            .with_name("Good Friday") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Easter Monday") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Early May Bank Holiday") \
            .on(first("monday").of("may")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("August Bank Holiday") \
            .on(last("monday").of("august")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Spring Bank Holiday") \
            .on(GB.date_of_spring_bank_holiday) \
            .with_flags("NV")

        """
        2023-05-08: Bank holiday for the coronation of King Charles III
        """
        self.define_holiday() \
            .with_name("Coronation of King Charles III") \
            .in_years([2023]) \
            .on(month=5, day=8) \
            .with_flags("NV")

        """
        2012-06-05: Queen's Diamond Jubilee
        """
        self.define_holiday() \
            .with_name("Queen's Diamond Jubilee") \
            .in_years([2012]) \
            .on(month=6, day=5) \
            .with_flags("NV")

        """
        2022-06-03: Queen's Platinum Jubilee
        """
        self.define_holiday() \
            .with_name("Queen's Platinum Jubilee") \
            .in_years([2022]) \
            .on(month=6, day=3) \
            .with_flags("NV")

        """
        2022-09-19: State Funeral of Queen Elizabeth II
        """
        self.define_holiday() \
            .with_name("State Funeral of Queen Elizabeth II") \
            .in_years([2022]) \
            .on(month=9, day=19) \
            .with_flags("NF")

    @staticmethod
    def date_of_spring_bank_holiday(year):
        """
        1. last monday in May
        2012: Moved to June 4, because of Queen’s Diamond Jubilee
        2022: Moved to June 2, because of Queen's Platinum Jubilee
        """
        if year == 2012:
            return SmartDayArrow(year, 6, 4)
        elif year == 2022:
            return SmartDayArrow(year, 6, 2)
        else:
            return last("monday").of("may")(year)

    @staticmethod
    def date_is_on_weekend(month, day):
        def wrapper(year):
            return SmartDayArrow(year, month, day).weekday() in ["saturday", "sunday"]

        return wrapper

    @staticmethod
    def date_is_on_saturday(month, day):
        def wrapper(year):
            return SmartDayArrow(year, month, day).weekday() == "saturday"

        return wrapper

    @staticmethod
    def date_is_on_sunday(month, day):
        def wrapper(year):
            return SmartDayArrow(year, month, day).weekday() == "sunday"

        return wrapper



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day, first


class SE(Country):
    id = "SE"
    languages = ["sv"]
    default_lang = "sv"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nyårsdagen") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Trettondedag jul") \
            .on(month=1, day=6) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Första maj") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Nationaldagen") \
            .on(month=6, day=6) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Julafton") \
            .on(month=12, day=24) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Juldagen") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Annandag jul") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Nyårsafton") \
            .on(month=12, day=31) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Långfredagen") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Påskdagen") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Annandag påsk") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Kristi himmelsfärdsdag") \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pingstdagen") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Midsommarafton") \
            .on(day(1).before(first("saturday").after(month=6, day=19))) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Midsommardagen") \
            .on(first("saturday").after(month=6, day=19)) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Alla helgons dag") \
            .on(first("saturday").after(month=10, day=30)) \
            .with_flags("NRV")



from holidata.holiday import Region


class BE(Region):
    """
    2020-05-09 -           : https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV7P1
    2019-02-07 - 2020-05-08: https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV6P1
    2018-01-01 - 2019-02-06: https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV5P1
    2015-10-25 - 2017-12-31: https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV4P1
    2010-12-29 - 2015-10-24: https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV2P1
    1995-01-01 - 2010-12-28: https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV2P1
    """
    def __init__(self, country):
        super().__init__("BE", country)

        """
        2020-05-08: 75. Jahrestag der Befreiung vom Nationalsozialismus und der Beendigung des Zweiten Weltkrieges in Europa
        Introduced 2019 for Berlin
        https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV6P1
        """
        self.define_holiday() \
            .with_name("75. Jahrestag der Befreiung vom Nationalsozialismus und der Beendigung des Zweiten Weltkrieges in Europa") \
            .in_years([2020]) \
            .on(month=5, day=8) \
            .with_flags("F")

        """
        03-08: Frauentag
        Introduced 2019 for Berlin
        https://gesetze.berlin.de/perma?d=jlr-FeiertGBEV6P1
        """
        self.define_holiday() \
            .with_name("Internationaler Frauentag") \
            .since(2019) \
            .on(month=3, day=8) \
            .with_flags("F")



from holidata.holiday import Region
from holidata.utils import SmartDayArrow, day


class MC(Region):
    def __init__(self, country):
        super().__init__("MC", country)

        self.define_holiday() \
            .with_name("Lunes siguiente al Año Nuevo") \
            .in_years([2017, 2023]) \
            .on(month=1, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Epifanía del Señor") \
            .in_years([2013, 2019]) \
            .on(month=1, day=7) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Día de la Región de Murcia") \
            .except_for([2013, 2024]) \
            .on(MC.dia_de_la_region_de_murcia) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("San José") \
            .in_years([2011, 2012, 2013, 2014, 2015, 2016, 2018, 2019, 2020, 2021, 2024]) \
            .on(month=3, day=19) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Lunes siguiente a La Inmaculada Concepción") \
            .in_years([2013, 2024]) \
            .on(month=12, day=9) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("Jueves Santo") \
            .on(day(3).before(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes de Pascua") \
            .in_years([2023]) \
            .on(day(1).after(country.easter())) \
            .with_flags("RV")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Fiesta del Trabajo") \
            .in_years([2011, 2022]) \
            .on(month=5, day=2) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente al Día de la Constitución Española") \
            .in_years([2015, 2020]) \
            .on(month=12, day=7) \
            .with_flags("F")

        self.define_holiday() \
            .with_name("Lunes siguiente a la Natividad del Señor") \
            .in_years([2022]) \
            .on(month=12, day=26) \
            .with_flags("RF")

        self.define_holiday() \
            .with_name("San Esteban") \
            .in_years([2016]) \
            .on(month=12, day=26) \
            .with_flags("RF")

    @staticmethod
    def dia_de_la_region_de_murcia(year):
        if year == 2019:
            return SmartDayArrow(year, 6, 10)
        else:
            return SmartDayArrow(year, 6, 9)



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import SmartDayArrow, day


class ZA(Country):
    id = "ZA"
    languages = ["en"]
    default_lang = "en"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Christmas Day") \
            .on(month=12, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Good Friday") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Family Day") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("New Year's Day") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("New Year's Day (Supplement)") \
            .on(month=1, day=2) \
            .with_flags("NF") \
            .annotated_with("Supplement holiday") \
            .on_condition(ZA.date_is_on_sunday(month=1, day=1))

        self.define_holiday() \
            .with_name("Human Rights Day") \
            .on(month=3, day=21) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Human Rights Day (Supplement)") \
            .on(month=3, day=22) \
            .with_flags("NF") \
            .annotated_with("Supplement holiday") \
            .on_condition(ZA.date_is_on_sunday(month=3, day=21))

        self.define_holiday() \
            .with_name("Freedom Day") \
            .on(month=4, day=27) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Freedom Day (Supplement)") \
            .on(month=4, day=28) \
            .with_flags("NF") \
            .annotated_with("Supplement holiday") \
            .on_condition(ZA.date_is_on_sunday(month=4, day=27))

        self.define_holiday() \
            .with_name("Worker's Day") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Worker's Day (Supplement)") \
            .on(month=5, day=2) \
            .with_flags("NF") \
            .annotated_with("Supplement holiday") \
            .on_condition(ZA.date_is_on_sunday(month=5, day=1))

        self.define_holiday() \
            .with_name("Youth Day") \
            .on(month=6, day=16) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Youth Day (Supplement)") \
            .on(month=6, day=17) \
            .with_flags("NF") \
            .annotated_with("Supplement holiday") \
            .on_condition(ZA.date_is_on_sunday(month=6, day=16))

        self.define_holiday() \
            .with_name("National Women's Day") \
            .on(month=8, day=9) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("National Women's Day (Supplement)") \
            .on(month=8, day=10) \
            .with_flags("NF") \
            .annotated_with("Supplement holiday") \
            .on_condition(ZA.date_is_on_sunday(month=8, day=9))

        self.define_holiday() \
            .with_name("Heritage Day") \
            .on(month=9, day=24) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Heritage Day (Supplement)") \
            .on(month=9, day=25) \
            .with_flags("NF") \
            .annotated_with("Supplement holiday") \
            .on_condition(ZA.date_is_on_sunday(month=9, day=24))

        self.define_holiday() \
            .with_name("Day of Reconciliation") \
            .on(month=12, day=16) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Day of Reconciliation (Supplement)") \
            .on(month=12, day=17) \
            .with_flags("NF") \
            .annotated_with("Supplement holiday") \
            .on_condition(ZA.date_is_on_sunday(month=12, day=16))

        self.define_holiday() \
            .with_name("Day of Goodwill") \
            .on(month=12, day=26) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Day of Goodwill (Supplement)") \
            .on(month=12, day=27) \
            .with_flags("NF") \
            .annotated_with("Supplement holiday") \
            .on_condition(ZA.date_is_on_sunday(month=12, day=26))

    @staticmethod
    def date_is_on_sunday(month, day):
        def wrapper(year):
            return SmartDayArrow(year, month, day).weekday() == "sunday"

        return wrapper



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import SmartDayArrow, day, first, fourth


class NZ(Country):
    id = "NZ"
    languages = ["en"]
    default_lang = "en"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("New Year's Day") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("New Year's Day (observed)") \
            .on(first("monday").after(month=1, day=1)) \
            .with_flags("NV") \
            .on_condition(NZ.date_is_on_weekend(month=1, day=1))

        self.define_holiday() \
            .with_name("Day after New Year's Day") \
            .on(NZ.day_after_new_years_day) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Waitangi Day") \
            .on(month=2, day=6) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Waitangi Day (observed)") \
            .since(2017) \
            .on(first("monday").after(month=2, day=6)) \
            .with_flags("NV") \
            .on_condition(NZ.date_is_on_weekend(month=2, day=6))

        self.define_holiday() \
            .with_name("ANZAC Day") \
            .on(month=4, day=25) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("ANZAC Day (observed)") \
            .since(2016) \
            .on(first("monday").after(month=4, day=25)) \
            .with_flags("NV") \
            .on_condition(NZ.date_is_on_weekend(month=4, day=25))

        self.define_holiday() \
            .with_name("Christmas Day") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Christmas Day (observed)") \
            .on(first("tuesday").after(month=12, day=25)) \
            .with_flags("NV") \
            .on_condition(NZ.date_is_sunday(month=12, day=25))

        self.define_holiday() \
            .with_name("Christmas Day (observed)") \
            .on(first("monday").after(month=12, day=25)) \
            .with_flags("NV") \
            .on_condition(NZ.date_is_saturday(month=12, day=25))

        self.define_holiday() \
            .with_name("Boxing Day") \
            .on(month=12, day=26) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Boxing Day (observed)") \
            .on(first("tuesday").after(month=12, day=26)) \
            .with_flags("NV") \
            .on_condition(NZ.date_is_sunday(month=12, day=26))

        self.define_holiday() \
            .with_name("Boxing Day (observed)") \
            .on(first("monday").after(month=12, day=26)) \
            .with_flags("NV") \
            .on_condition(NZ.date_is_saturday(month=12, day=26))

        self.define_holiday() \
            .with_name("Good Friday") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Easter Monday") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Queen's Birthday") \
            .on(first("monday").of("june")) \
            .with_flags("NV")

        self.define_holiday() \
            .with_name("Labour Day") \
            .on(fourth("monday").of("october")) \
            .with_flags("NV")

    @staticmethod
    def date_is_on_weekend(month, day):
        def wrapper(year):
            return SmartDayArrow(year, month, day).weekday() in ["saturday", "sunday"]

        return wrapper

    @staticmethod
    def date_is_saturday(month, day):
        def wrapper(year):
            return SmartDayArrow(year, month, day).weekday() == "saturday"

        return wrapper

    @staticmethod
    def date_is_sunday(month, day):
        def wrapper(year):
            return SmartDayArrow(year, month, day).weekday() == "sunday"

        return wrapper

    @staticmethod
    def day_after_new_years_day(year):
        date = SmartDayArrow(year, 1, 2)

        if date.weekday() in ["sunday", "monday"]:
            return date.shift_to_weekday("tuesday", including=True)

        elif date.weekday() == "saturday":
            return date.shift_to_weekday("monday", including=True)

        return date



import csv
import io
import json
from typing import List, Dict, Any, Callable

from holidata.holiday import Holiday
from holidata.plugin import PluginMount


class Emitter(metaclass=PluginMount):
    type: str = None

    def __init__(self):
        if self.type is None:
            raise ValueError(f"Emitter {self.__class__.__name__} does not provide its type!")

    @staticmethod
    def get(identifier: str) -> Callable[[], 'Emitter']:
        return Emitter.get_plugin(identifier, "type")

    def output(self, holidays: List[Holiday]) -> str:
        raise NotImplementedError


class JsonEmitter(Emitter):
    type: str = "json"

    def output(self, holidays: List[Holiday]) -> str:
        export_data: List[Dict[str, Any]] = [h.as_dict() for h in holidays]
        export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))
        return "\n".join([json.dumps(h, ensure_ascii=False, sort_keys=False, indent=None, separators=(",", ":")) for h in export_data]) + "\n"


class CsvEmitter(Emitter):
    type: str = "csv"

    def output(self, holidays: List[Holiday]) -> str:
        export_data: List[Dict[str, Any]] = [h.as_dict() for h in holidays]
        export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))
        result: io.StringIO = io.StringIO()

        writer: csv.DictWriter = csv.DictWriter(
            result,
            ["locale", "region", "date", "description", "type", "notes"],
            quoting=csv.QUOTE_ALL,
            lineterminator="\n")
        writer.writeheader()
        writer.writerows(export_data)

        return result.getvalue()


class YamlEmitter(Emitter):
    type: str = "yaml"

    @staticmethod
    def _format_yaml(holiday: Dict[str, Any]) -> str:
        output: str = "  holiday:\n"
        for key in ["locale", "region", "date", "description", "type", "notes"]:
            value: Any = holiday[key]

            if value is not None and value != "":
                output += f"    {key}: {value}\n"
            else:
                output += f"    {key}:\n"

        return output

    def output(self, holidays: List[Holiday]) -> str:
        export_data: List[Dict[str, Any]] = [h.as_dict() for h in holidays]
        export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))

        output: str = "%YAML 1.1\n"
        output += "---\n"

        for holiday in export_data:
            output += YamlEmitter._format_yaml(holiday)

        output += "...\n"
        return output


class XmlEmitter(Emitter):
    type: str = "xml"

    @staticmethod
    def _format_xml(holiday: Dict[str, Any]) -> str:
        output: str = "  <holiday>\n"

        for key in ["locale", "region", "date", "description", "type", "notes"]:
            value: Any = holiday[key] if key in holiday else ""
            output += f"    <{key}>{value if value is not None else ''}</{key}>\n"

        output += "  </holiday>\n"
        return output

    def output(self, holidays: List[Holiday]) -> str:
        export_data: List[Dict[str, Any]] = [h.as_dict() for h in holidays]
        export_data.sort(key=lambda x: (x["date"], x["description"], x["region"]))

        output: str = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n"
        output += "<holidays>\n"

        for holiday in export_data:
            output += XmlEmitter._format_xml(holiday)

        output += "</holidays>\n"
        return output



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day

"""
source: http://prawo.sejm.gov.pl/isap.nsf/download.xsp/WDU20150000090/O/D20150090.pdf
"""


class PL(Country):
    id = "PL"
    languages = ["pl"]
    default_lang = "pl"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nowy Rok") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Trzech Króli") \
            .on(month=1, day=6) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Święto Pracy") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Święto Konstytucji Trzeciego Maja") \
            .on(month=5, day=3) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Wniebowzięcie Najświętszej Maryi Panny") \
            .on(month=8, day=15) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Wszystkich Świętych") \
            .on(month=11, day=1) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Narodowe Święto Niepodległości") \
            .on(month=11, day=11) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Boże Narodzenie (pierwszy dzień)") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Boże Narodzenie (drugi dzień)") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Wielkanoc") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Poniedziałek Wielkanocny") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Zielone Świątki") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Boże Ciało") \
            .on(day(60).after(self.easter())) \
            .with_flags("NRV")



#!/usr/bin/env python3

import pathlib

from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="holidata",
    version="2024.8.2",
    description="Holidata is a utility for algorithmically producing holidays for a given locale and year",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/GothenburgBitFactory/holidata",
    author="Gothenburg Bit Factory",
    author_email="support@gothenburgbitfactory.org",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="holiday, calendar",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    python_requires=">=3.8, <4",
    install_requires=[
        "arrow >= 1.3.0",
        "docopt >= 0.6.2",
    ],
    extras_require={
        "test": [
            "pytest >= 7.4.2",
            "syrupy >= 4.6.1",
        ]
    },
    scripts=["bin/holidata"],
    project_urls={
        "Bug Reports": "https://github.com/GothenburgBitFactory/holidata/issues",
        "Source": "https://github.com/GothenburgBitFactory/holidata",
        "Website": "https://holidata.net",
    },
)



import datetime
from dataclasses import dataclass
from typing import Iterator, Callable, Union, Dict, List

import dateutil

from holidata.plugin import PluginMount
from holidata.utils import SmartDayArrow, date


@dataclass
class Holiday:
    """
    A sheer container for one holiday.
    """
    locale: str
    region: str
    date: datetime.date
    description: str
    flags: str = ""
    notes: str = ""

    def as_dict(self) -> dict:
        return {
            "locale": self.locale,
            "region": self.region,
            "date": self.date.strftime("%Y-%m-%d"),
            "description": self.description,
            "type": self.flags,
            "notes": self.notes
        }


class HolidayGenerator:
    def __init__(self, country_id: str, default_lang: str):
        self.name_dict: Dict[str, str] = {}
        self.notes: str = ""
        self.regions: List[str] = [""]
        self.flags: str = ""
        self.date: callable = None
        self.country_id: str = country_id
        self.default_lang: str = default_lang
        self.filters: List[callable] = []

    def with_name(self, name: str, lang: str = None) -> 'HolidayGenerator':
        lang = self.default_lang if lang is None else lang
        self.name_dict[lang] = name
        return self

    def with_names(self, name_dict: Dict[str, str]) -> 'HolidayGenerator':
        self.name_dict = name_dict
        return self

    def on(self, date_func: callable = None, month: int = None, day: int = None) -> 'HolidayGenerator':
        if month is not None and day is not None:
            self.date = date(month, day)
        elif callable(date_func):
            self.date = date_func
        else:
            raise ValueError("Invalid reference date")

        return self

    def with_flags(self, flags: str) -> 'HolidayGenerator':
        self.flags = flags
        return self

    def in_regions(self, regions: List[str]) -> 'HolidayGenerator':
        self.regions = regions
        return self

    def in_region(self, region: str) -> 'HolidayGenerator':
        self.regions = [region]
        return self

    def annotated_with(self, note: str) -> 'HolidayGenerator':
        self.notes = note
        return self

    def in_years(self, reference: List[int]) -> 'HolidayGenerator':
        def reference_does_contain(year: int) -> bool:
            return year in reference

        self.filters.append(reference_does_contain)
        return self

    def since(self, reference: int) -> 'HolidayGenerator':
        def reference_is_less_or_equal_to(year: int) -> bool:
            return reference <= year

        self.filters.append(reference_is_less_or_equal_to)
        return self

    def until(self, reference: int) -> 'HolidayGenerator':
        def reference_is_greater_or_equal_to(year: int) -> bool:
            return reference >= year

        self.filters.append(reference_is_greater_or_equal_to)
        return self

    def except_for(self, reference: List[int]) -> 'HolidayGenerator':
        def reference_does_not_contain(year: int) -> bool:
            return year not in reference

        self.filters.append(reference_does_not_contain)
        return self

    def on_condition(self, condition: callable) -> 'HolidayGenerator':
        self.filters.append(condition)
        return self

    def build_for_year_and_lang(self, year: int, lang: str) -> Iterator[Holiday]:
        for filter in self.filters:
            if not filter(year):
                return []

        for region in self.regions:
            date = self.date(year)
            if date is None:
                continue
            yield Holiday(
                locale=f"{lang}-{self.country_id}",
                region=region,
                date=date,
                description=self.name_dict[lang],
                flags=self.flags,
                notes=self.notes,
            )


class Country(metaclass=PluginMount):
    """
    Represents holidays of a given country
    """
    id: str
    languages: List[str]
    default_lang: Union[str, None] = None
    easter_type: str
    holiday_generators: list
    regions: list

    def __init__(self):
        if not self.id:
            raise ValueError(f"Country '{self.__class__.__name__}' does not provide an id!")
        if not self.languages:
            raise ValueError(f"Country '{self.__class__.__name__}' does not list any languages!")
        if self.default_lang is not None and self.default_lang not in self.languages:
            raise ValueError(f"Country '{self.__class__.__name__}' does not list language '{self.default_lang}'!")

        self.holiday_generators = []
        self.regions = []

    @staticmethod
    def get(identifier):
        return Country.get_plugin(identifier, "id")

    def validate_language_or_get_default(self, lang_id: str) -> str:
        if lang_id and lang_id.lower() not in self.languages:
            raise ValueError(
                f"Language '{lang_id}' is not defined for country '{self.id}'! Choose one of [{', '.join(self.languages)}].")
        elif lang_id is None and self.default_lang:
            lang_id = self.default_lang
        elif lang_id is None:
            raise ValueError(
                f"Country '{self.id}' has no default language specified! Choose one of [{', '.join(self.languages)}].")

        return lang_id.lower()

    def define_holiday(self) -> HolidayGenerator:
        generator = HolidayGenerator(self.id, self.default_lang)
        self.holiday_generators.append(generator)
        return generator

    def get_holidays_of(self, year: int, lang: str) -> Iterator[Holiday]:
        for generator in self.holiday_generators:
            yield from generator.build_for_year_and_lang(year, lang)
        for region in self.regions:
            yield from region.get_holidays_of(year, lang)

    def easter(self) -> Callable[[int], SmartDayArrow]:
        def wrapper(year: int) -> SmartDayArrow:
            date = dateutil.easter.easter(year, self.easter_type)
            return SmartDayArrow(date.year, date.month, date.day)

        return wrapper


class Region:
    def __init__(self, id: str, country: Country):
        self.id: str = id
        self.country: Country = country
        self.holiday_generators: List[HolidayGenerator] = []

    def define_holiday(self) -> HolidayGenerator:
        generator = HolidayGenerator(self.country.id, self.country.default_lang)
        generator.in_region(self.id)
        self.holiday_generators.append(generator)
        return generator

    def get_holidays_of(self, year: int, lang: str) -> Iterator[Holiday]:
        for generator in self.holiday_generators:
            yield from generator.build_for_year_and_lang(year, lang)



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.utils import day


class DK(Country):
    id = "DK"
    languages = ["da"]
    default_lang = "da"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.define_holiday() \
            .with_name("Nytårsdag") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Grundlovsdag") \
            .on(month=6, day=5) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Juledag") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Anden juledag") \
            .on(month=12, day=26) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Skærtorsdag") \
            .on(day(3).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Langfredag") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Påskedag") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Anden påskedag") \
            .on(day(1).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Store bededag") \
            .on(day(26).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Kristi himmelfartsdag") \
            .on(day(39).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pinsedag") \
            .on(day(49).after(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Anden pinsedag") \
            .on(day(50).after(self.easter())) \
            .with_flags("NRV")



from typing import Iterator

from holidata.emitters import Emitter
from holidata.holiday import Country, Holiday
from holidata.holidays import *


class Holidata:
    def __init__(self, holidays, emitter: Emitter = None):
        self.holidays = holidays
        self.emitter = emitter

    def __str__(self) -> str:
        return self.emitter.output(self.holidays)

    def formatted_as(self, format_id: str) -> 'Holidata':
        self.emitter = get_emitter_for(format_id)

        return self


class Locale:
    def __init__(self, country: Country, lang: str):
        self.country = country
        self.lang = lang

    def get_holidays_of(self, year: int) -> Iterator[Holiday]:
        return self.country.get_holidays_of(year, self.lang)

    def holidays_of(self, year: str) -> Holidata:
        return Holidata(self.country.get_holidays_of(Locale._parse_year(year), self.lang))

    @staticmethod
    def _parse_year(year: str) -> int:
        try:
            return int(year)
        except ValueError:
            raise ValueError(f"Invalid year '{year}'! Has to be an integer.")

    @property
    def id(self) -> str:
        return f"{self.lang}-{self.country.id}"


def get_country_for(identifier: str) -> Country:
    country_class = Country.get(identifier)

    if not country_class:
        raise ValueError(f"No country found for id '{identifier}'!")

    return country_class()


def get_emitter_for(identifier: str) -> Emitter:
    emitter_class = Emitter.get(identifier)

    if not emitter_class:
        raise ValueError(f"Unsupported output format '{identifier}'!")

    return emitter_class()


def for_locale(country_id: str, lang_id: str = None) -> Locale:
    country = get_country_for(country_id)
    lang_id = country.validate_language_or_get_default(lang_id)

    return Locale(country, lang_id)



from dateutil.easter import EASTER_WESTERN

from holidata.holiday import Country
from holidata.holidays.ES.AN import AN
from holidata.holidays.ES.AR import AR
from holidata.holidays.ES.AS import AS
from holidata.holidays.ES.CB import CB
from holidata.holidays.ES.CE import CE
from holidata.holidays.ES.CL import CL
from holidata.holidays.ES.CM import CM
from holidata.holidays.ES.CN import CN
from holidata.holidays.ES.CT import CT
from holidata.holidays.ES.EX import EX
from holidata.holidays.ES.GA import GA
from holidata.holidays.ES.IB import IB
from holidata.holidays.ES.MC import MC
from holidata.holidays.ES.MD import MD
from holidata.holidays.ES.ML import ML
from holidata.holidays.ES.NC import NC
from holidata.holidays.ES.PV import PV
from holidata.holidays.ES.RI import RI
from holidata.holidays.ES.VC import VC
from holidata.utils import day

__all__ = [
    "ES"
]

"""
Information taken from government websites around 2020-06
    https://administracion.gob.es/pag_Home/atencionCiudadana/calendarios/laboral.html
    http://www.seg-social.es/wps/portal/wss/internet/CalendarioLaboral

    2011: https://www.boe.es/eli/es/res/2010/10/07/(1)
          https://www.boe.es/eli/es/res/2010/11/24/(1)
    2012: https://www.boe.es/eli/es/res/2011/10/06/(1)
    2013: https://www.boe.es/eli/es/res/2012/10/30/(1)
          https://www.boe.es/eli/es/res/2012/11/12/(2)
    2014: https://www.boe.es/eli/es/res/2013/11/08/(3)
          https://www.boe.es/eli/es/res/2013/11/21/(1)
    2015: https://www.boe.es/eli/es/res/2014/10/17/(3)
    2016: https://www.boe.es/eli/es/res/2015/10/19/(1)
    2017: https://www.boe.es/eli/es/res/2016/10/04/(1)
    2018: https://www.boe.es/eli/es/res/2017/10/09/(1)
          https://www.boe.es/eli/es/res/2017/10/09/(1)/corrigendum/20171019
          https://www.boe.es/eli/es/res/2017/10/09/(1)/corrigendum/20171025
    2019: https://www.boe.es/eli/es/res/2018/10/16/(1)
    2020: https://www.boe.es/eli/es/res/2019/10/03/(1)
    2021: https://www.boe.es/eli/es/res/2020/10/28/(1)
    2022: https://www.boe.es/eli/es/res/2021/10/14/(3)
    2023: https://www.boe.es/eli/es/res/2022/10/07/(2)
    2024: https://www.boe.es/eli/es/res/2023/10/23/(1)

Regional governments
    [AN] https://www.juntadeandalucia.es/temas/trabajar/relaciones/calendario.html

Also those sites for some information
    https://es.wikipedia.org/wiki/Calendario_laboral
"""


class ES(Country):
    id = "ES"
    languages = ["es"]
    default_lang = "es"
    easter_type = EASTER_WESTERN

    def __init__(self):
        super().__init__()

        self.regions = [
            AN(self),
            AR(self),
            AS(self),
            CB(self),
            CE(self),
            CL(self),
            CM(self),
            CN(self),
            CT(self),
            EX(self),
            GA(self),
            IB(self),
            MC(self),
            MD(self),
            ML(self),
            NC(self),
            PV(self),
            RI(self),
            VC(self),
        ]

        self.define_holiday() \
            .with_name("Año Nuevo") \
            .on(month=1, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Epifanía del Señor") \
            .on(month=1, day=6) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Fiesta del Trabajo") \
            .on(month=5, day=1) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Asunción de la Virgen") \
            .on(month=8, day=15) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Fiesta Nacional de España") \
            .on(month=10, day=12) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Todos los Santos") \
            .on(month=11, day=1) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Día de la Constitución Española") \
            .on(month=12, day=6) \
            .with_flags("NF")

        self.define_holiday() \
            .with_name("Inmaculada Concepción") \
            .on(month=12, day=8) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Natividad del Señor") \
            .on(month=12, day=25) \
            .with_flags("NRF")

        self.define_holiday() \
            .with_name("Viernes Santo") \
            .on(day(2).before(self.easter())) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Pascua") \
            .on(self.easter()) \
            .with_flags("NRV")

        self.define_holiday() \
            .with_name("Lunes siguiente al Año Nuevo") \
            .in_years([2012]) \
            .on(month=1, day=2) \
            .with_flags("NF")
